package org.objectweb.dsrg.sofa.examples.logaspect;

import org.objectweb.dsrg.sofa.SOFAException;
import java.util.Map;
import org.objectweb.dsrg.sofa.microarchitecture.ComponentDescription;
import org.objectweb.dsrg.sofa.microarchitecture.SOFAMicroComponent;
import org.objectweb.dsrg.sofa.bootstrap.MIInterceptor;
/** Micro log transceiver.
  *
  * Listens notifications from interceptors and send logs to MLogger.
  *
  * @author Ondrej Cerny
  */
public class MLogTransceiver implements SOFAMicroComponent {
  
  /** Notification receiver.
    * 
    * <p>Utilized for receiving notifications from interceptor on
    * appropriate interface of MLogTransceiver and sends logs
    * to MLogTransceiver (via its sendLog() method) which
    * forwards the log to MLogger (via log interface).</p>
    *
    * <p>This class also contains information about the context
    * of notification (component, caller/callee, bussines/control)
    * which are added to the log.</p>
    */
  private class Receiver implements MIInterceptor {
    private MLogTransceiver sender;
    private MILog.View view;
    private MILog.Type type;
    private final MILog.Event callEvent = MILog.Event.CALL;
    private final MILog.Event returnEvent = MILog.Event.RETURN;

    /** Constructor. 
      * 
      * @param sender object responsible for sending logs
      * @param type log event type
      * @param view observation point of view
      */
    public Receiver(MLogTransceiver sender, MILog.Type type, MILog.View view) {
      this.sender = sender;
      this.view = view;
      this.type = type;
    }
	  /** Receives notification from interceptor and passes the log to the sender.
      *
      * @param method textual method representation
	    */
  	public void notifyCall(String method) {
	  	sender.sendLog(callEvent, type, view, method);
  	}

	  /** Receives notification from interceptor and passes the log to the sender.
      *
      * @param method textual method representation
	    */
  	public void notifyReturn(String method) {
	  	sender.sendLog(returnEvent, type, view, method);
  	}
  }

  private String componentDescription = null;
  private MILog logger = null;
  private Receiver receiverRequired = null;
  private Receiver receiverProvided = null;
  private Receiver receiverControl = null;

	/** Sends the log to the required interface.
	  * 
    * @param event class of logged event
    * @param type type of logged event
    * @param view observation point of view 
    * @param method logged method (textual description)
	  */
	protected void sendLog(MILog.Event event, MILog.Type type, MILog.View view, String method) {
		if (logger != null) {
      logger.log((componentDescription == null) ? "<unknown>" : componentDescription,
                 event, type, view, (method == null) ? "<unknown>" : method);
		}
	}

  /* SOFAMicroComponent's methods. */

	/** Initialize this component.
	  * 
	  */
	public void init(ComponentDescription component, Map<String, String> parameters) throws SOFAException {
		Object content = (component == null) ? null : component.getComponentContent();
    if (content == null) {
      componentDescription = "UNKNOWN(composite component)";
    } else {
      componentDescription = content.toString();
    }
  }
	
	/** Returns appropriate {@link Receiver} as provided interface.
    *
    * <p>Creates (or fetches) Receiver with parameters appropriate
    * to the interface (depending on <code>name</code> parameter).
    *
    * @param name name of the interface
    * @return Receiver instance
	  */
	public Object getProvided(String name) throws SOFAException {
    MIInterceptor result = null;
		if (name == null) {
			throw new SOFAException("The microcomponent does not provide <null> interface.");
		}
    if (name.equals("listen-required")) {
      if (receiverRequired == null) {
        receiverRequired = new Receiver(this, MILog.Type.BUSSINESS, MILog.View.CALLER);
      }
      result = (MIInterceptor)receiverRequired;
    } else if (name.equals("listen-provided")) {
      if (receiverProvided == null) {
        receiverProvided = new Receiver(this, MILog.Type.BUSSINESS, MILog.View.CALLEE);
      }
      result = (MIInterceptor)receiverProvided;
    } else if (name.equals("listen-control")) {
      if (receiverControl == null) {
        receiverControl = new Receiver(this, MILog.Type.CONTROL, MILog.View.UNSPECIFIED);
      }
      result = (MIInterceptor)receiverControl;
    } else {
			throw new SOFAException("The microcomponent does not provide interface called '" + name + "'.");
    }
    return result;
	}

	/** Sets required interface (log).
	  * 
	  */
	public void setRequired(String name, Object ref) throws SOFAException {
		if (name != null && name.equals("log")) {
			if (ref instanceof MILog) {
			    logger = (MILog)ref;
			} else {
			    throw new SOFAException("Unable to set required interface 'log'. Incompatible type.");
			}
   	}
		// we silently ignore interfaces we dont need
	}
}
