/* $Id: StreamConsumerImpl.java 278 2008-08-15 18:17:30Z mmalohlava $
 *
 * SOFA 2.0 component system
 * Copyright (C) 2006-2008  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.examples.streaming.arch;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import org.objectweb.dsrg.connector.streaming.StreamProvider;
import org.objectweb.dsrg.sofa.SOFAClient;
import org.objectweb.dsrg.sofa.SOFALifecycle;
import org.objectweb.dsrg.sofa.examples.streaming.iface.MediaServer;

/**
 * Client side of the stream example.
 * 
 * The client has to references to the servers - one is media server providing information about
 * content, the second one is a stream server providing access to the streamed media.
 * 
 * @author Michal Malohlava
 *
 */
public class StreamConsumerImpl implements SOFAClient, SOFALifecycle, Runnable {
	
	/** Reference to content info provider */
	private MediaServer mediaServer;
	
	/** Reference to stream provider */
	private StreamProvider streamProvider;
	
	/** Quit the running thread */
	private boolean quitThread = false;
	
	/** Thread for this component */
	private Thread thread;
	
	/** default output file */
	public static final String OUTPUT_FILENAME_PREFIX = "local_";
	
	// -----------------------------------
	// SOFAClient interface implementation
	// -----------------------------------

	@Override
	public void setRequired(String name, Object iface) {
		if ("media-server-iface".equals(name)) {
			if (iface instanceof MediaServer) {
				mediaServer = (MediaServer) iface;
			}
		}
		
		if ("stream-provider-iface".equals(name)) {
			if (iface instanceof StreamProvider) {
				streamProvider = (StreamProvider) iface; 
			}
		}		
		
	}

	// --------------------------------------
	// SOFALifeCycle interface implementation
	// -------------------------------------
	@Override
	public void start() {
		thread = new Thread(this);
		
		thread.start();		
	}
	
	@Override
	public void stop() {
		quitThread = true;
				
		thread.interrupt();		
		try {
			thread.join();
		} catch (InterruptedException e) {
			// FIXME join again or continue ?
			e.printStackTrace();
		}
	}

	@Override
	public void stopping() {		
	}

	// Runnable iface methods
	@Override
	public void run() {
		while (!quitThread) {
			String selectedId = null;
			
			// test code			
			String[] mediaList = mediaServer.getContentList();
			System.out.println("[client compo] Media list");
			for (int i = 0; i < mediaList.length; i++ ) {
				System.out.println(mediaList[i]);
				// select one mp3 and play it
				if (mediaList[i].endsWith(".mp3")) {
					selectedId = mediaList[i];
				}
			}
			
			if (selectedId != null) {
				System.err.println("[client compo] Getting media with id = " + selectedId);
				System.err.println("[client compo] Media info:" + mediaServer.getContentInfo(selectedId));
				
				InputStream is = streamProvider.getStream(selectedId);
				try {
					handleRemoteStream(selectedId, is);
				} catch (IOException e) {
					System.err.println("[client compo] Error occured during handling remote input stream");
					e.printStackTrace();
				} finally {
					try {					
						is.close();
					} catch (IOException e) {
						System.err.println("[client compo] Error occured during closing input stream received from server");
						e.printStackTrace();
					} 
				}
			} else {
				System.out.println("[client compo] No stream source selected.");
			}
			
			try {
				Thread.sleep(5000);
			} catch (InterruptedException e) {
				if (quitThread)
					break;
			}			
			
		}		
	}
	
	protected void handleRemoteStream(String id, InputStream is) throws IOException {
		String filename = OUTPUT_FILENAME_PREFIX + id;
		OutputStream os = new FileOutputStream(filename);
		int bufferSize = 4096;
		
		System.err.println("[client compo] Writing remote intput stream into file " + filename);
		byte buffer[] = new byte[bufferSize];
		int res = 0;
		while ((res = is.read(buffer, 0, bufferSize )) != -1) {
			os.write(buffer, 0, res);
		}
		
		os.close();
	}
	
}
