/* $Id$
 *
 * SOFA 2.0 component system
 * Copyright (C) 2006-2010  DDDS, Charles University in Prague
 *                          http://d3s.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.sofa.aspects.script;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.script.ScriptEngineFactory;
import javax.script.ScriptEngineManager;

import org.apache.log4j.Logger;

/**
 * Manager class for creation of the {@link EngineInitializer} instances.
 * 
 * Prepares instances for all available scripting languages and their aliases.
 * 
 * The list of supported languages is expanded using the
 * {@link ScriptEngineManager} to contain all language aliases.
 * 
 * @author Jaroslav Keznikl
 */
public class EngineInitializerManager {

	final static Logger logger = Logger.getLogger(EngineInitializerManager.class);
	
	/** Associated script engine manager */
	ScriptEngineManager scriptManager;
	
	/**
	 * Map of all {@link EngineInitializer} instances for all supported
	 * languages ant their aliases.
	 */
	Map<String, EngineInitializer> initializers;
	
	/**
	 * Map of default {@link EngineInitializer} instances for their default languages.
	 */
	Map<String, EngineInitializer> defaultInitializers;
	
	/**
	 * Initializes the manager and expands the language aliases.
	 * @param scriptManager The associated script engine manager.
	 */
	public EngineInitializerManager(ScriptEngineManager scriptManager) {
		this.scriptManager = scriptManager;
		init(scriptManager);
	}

	/**
	 * Returns an {@link EngineInitializer} instance according to the given
	 * language (or its alias).
	 * 
	 * The aliases are taken from the {@link ScriptEngineManager}.
	 * 
	 * @param language The target language.
	 * @return The associated {@link EngineInitializer} instance.
	 */
	public EngineInitializer getInitializerByLanguage(String language) {
		if (initializers.containsKey(language.toLowerCase()))
			return initializers.get(language.toLowerCase());
		else
			return null;
	}
	
	/**
	 * Returns, whether the given implementation language is supported. 
	 * @param language	the queried implementation language
	 * @return	true if the given implementation language is supported
	 */
	public boolean supportsLanguage(String language) {
		return initializers.containsKey(language.toLowerCase());
	}

	/**
	 * Initialize the manager and expand the language aliases.
	 * 
	 * Initializes the default initializers map and then expands this map for
	 * all language aliases using the {@link ScriptEngineManager}.
	 * 
	 * @param mgr The associated script engine manager.
	 */
	void init(ScriptEngineManager mgr) {
		initializers = new HashMap<String, EngineInitializer>();
		defaultInitializers = new HashMap<String, EngineInitializer>();	
		
		defaultInitializers.put(
				PythonEngineInitializer.LANGUAGE_NAME, 
				new PythonEngineInitializer(mgr));
		defaultInitializers.put(
				JavascriptEngineInitializer.LANGUAGE_NAME, 
				new JavascriptEngineInitializer(mgr));
		
		List<ScriptEngineFactory> factories = mgr.getEngineFactories();
		for (ScriptEngineFactory factory : factories) {
			String langName = factory.getLanguageName();
			
			logger.trace("Processing factory for language: " + langName);
			
			// skip unsupported languages
			if (!defaultInitializers.containsKey(langName))
				continue;
			
			EngineInitializer initializer = defaultInitializers.get(langName);			
			
			List<String> engNames = factory.getNames();
			for (String name : engNames) {
				logger.trace("Adding initializer for name: " + name);
				initializers.put(name.toLowerCase(), initializer);
			}			
		}
	}
}
