/* $Id$
 *
 * SOFA 2.0 component system
 * Copyright (C) 2006-2010  DDDS, Charles University in Prague
 *                          http://d3s.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.sofa.aspects.script;

import org.apache.log4j.Logger;
import org.objectweb.dsrg.sofa.SOFAClassLoader;
import org.objectweb.dsrg.sofa.SOFAException;

import java.lang.reflect.Array;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.script.ScriptException;

import org.objectweb.dsrg.sofa.microarchitecture.ComponentDescription;
import org.objectweb.dsrg.sofa.microarchitecture.SOFAMicroComponent;
import org.objectweb.dsrg.sofa.repository.model.Interface;
import org.objectweb.dsrg.sofa.repository.model.InterfaceType;
import org.objectweb.dsrg.sofa.bootstrap.MIInterceptor;
import org.ow2.dsrg.sofa.aspects.script.NotScriptException;

/**
 * Script call transceiver.
 * 
 * Listens notifications from script interceptors and send logs to the
 * {@link MIScriptInvoker}.
 * 
 * Contains basic implementation of the declared exception handling (not yet supported).
 * 
 * @author Jaroslav Keznikl
 */
public class MCallTransceiver implements SOFAMicroComponent, MIScriptInterceptor {
	
	final static private Logger logger = Logger.getLogger(MCallTransceiver.class);
	final private SOFAClassLoader loader = SOFAClassLoader.getSOFAClassLoader();
	
	/** Required interface for transmition of the interceptor calls. */
	private MIScriptInvoker invoker = null;
	
	/** Map of the declared exceptions of all provided interfaces. */
	private Map<String, Map<String, List<Class>>> interfaces = new HashMap<String, Map<String, List<Class>>>();

	/** Transmits the call from the {@link MIScriptInterceptor} to the {@link MIScriptInvoker}.
	 * @param methodSignature	Signature of the called method.
	 * @param interfaceClass	Bytecode signature of the associated interface.
	 * @param params	Parameters of the call.  
	 * @return Value returned from the {@link MIScriptInvoker}.
	 * @throws Exception	Throws {@link SOFAException} if an error occurs.
	 */
	public Object notifyCall(String methodSignature, String interfaceClass, Object[] params)
			throws Exception {
		if (invoker == null) 
			throw new SOFAException("The required interface 'call' is not set!");
			
		String methodName = methodSignature.substring(0, methodSignature.indexOf("("));
	
		try {
			return invoker.invoke(methodName, params);
		} catch (NoSuchMethodException e) {			
			throw new SOFAException(e);
		} catch (ScriptException e) {			
			logger.trace("Catched ScriptException " + e.toString() + " with cause " + e.getCause());
			throw new SOFAException(e);
		}
	}
	
	/**
	 * Initialize this component.
	 * 
	 * <p>Search all methods of all provided interfaces for declared exceptions
	 * and store them in the the {@link #interfaces} map.</p>
	 */
	public void init(ComponentDescription component, Map<String, String> parameters) throws SOFAException {
		Object content = (component == null) ? null : component
				.getComponentContent();
		
		// search all provided interfaces
		for (Interface iface : component.getComponentFrame().getProvidedInterface()) {
			InterfaceType ifaceType = iface.getInterfaceType();
			try {
				Class ifaceClass = loader.loadClass(ifaceType.getCodeBundle(), ifaceType.getSignature());
				String ifaceDesc = ScriptHelper.classToDesc(ifaceClass);
				Map<String, List<Class>> methodMap = new HashMap<String, List<Class>>();
				
				interfaces.put(ifaceDesc, methodMap);
				
				// process only public methods
				for (Method method : ifaceClass.getMethods()) {
					String desc = ScriptHelper.methodToDesc(method);
					List<Class> exceptions = new ArrayList<Class>();
 
					if (methodMap.containsKey(desc))
						exceptions = methodMap.get(desc);						
					else 
						exceptions = new ArrayList<Class>();
					
					// add all declared exceptions
					for (Class ex: method.getExceptionTypes())
						exceptions.add(ex);
					
					methodMap.put(desc, exceptions);
				}				
			} catch (ClassNotFoundException e) {
				throw new SOFAException("Frame interface class not found!", e);
			}
		}
	}
	


	/**
	 * Returns this reference for the listen-provided interface. 
	 */
	public Object getProvided(String name) throws SOFAException {		
		if (name == null) {
			throw new SOFAException("The microcomponent does not provide <null> interface.");
		}
		if (name.equals("listen-provided")) {
			return this;
		} else {
			throw new SOFAException("The microcomponent does not provide interface called '" + name + "'.");
		}		
	}

	/**
	 * Sets required the interface (invoker).
	 */
	public void setRequired(String name, Object ref) throws SOFAException {		
		if (name != null && name.equals("invoker")) {
			if (ref instanceof MIScriptInvoker) {
				invoker = (MIScriptInvoker) ref;
			} else {
				throw new SOFAException(
						"Unable to set required interface 'call'. Incompatible type.");
			}
		}
		// we silently ignore interfaces we dont need
	}

	
	
}
