/* $Id$
 *
 * SOFA 2.0 component system
 * Copyright (C) 2006-2010  DDDS, Charles University in Prague
 *                          http://d3s.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.sofa.aspects.script;

import org.objectweb.dsrg.sofa.SOFAException;
import java.util.Map;
import org.objectweb.dsrg.sofa.microarchitecture.ComponentDescription;
import org.objectweb.dsrg.sofa.microarchitecture.LifecycleState;
import org.objectweb.dsrg.sofa.microarchitecture.MILifecycle;
import org.objectweb.dsrg.sofa.microarchitecture.SOFAMicroComponent;
import org.objectweb.dsrg.sofa.microarchitecture.SOFAMicroInterceptor;

/**
 * Implementation of microcomponent that intercepts the {@link MILifecycle} control interface. 
 * 
 * Redirects required requests to the {@link MScriptInvoker} microcomponent.
 * 
 * @author Jaroslav Keznikl
 */
public class MLifecycleProxy implements SOFAMicroComponent, SOFAMicroInterceptor, MILifecycle {
  
	/** The required interface reference. */
	MIScriptInvoker scriptInvoker;
	/** The next element in the control interface's delegation chain. */
	MILifecycle endpoint;
	
	public void init(ComponentDescription component,
			Map<String, String> parameters) throws SOFAException {		
	}

	/** No provided interfaces available. */
    public Object getProvided (String name) throws SOFAException {
    	return null;
    }

	/**
	 * Sets the required interface (invoker).
	 */
	public void setRequired(String name, Object ref) throws SOFAException {		
		if (name != null && name.equals("invoker")) {
			if (ref instanceof MIScriptInvoker) {
				scriptInvoker = (MIScriptInvoker) ref;
			} else {
				throw new SOFAException(
						"Unable to set required interface 'lifecycle'. Incompatible type.");
			}
		}
		// we silently ignore interfaces we don't need
	}

	
	public Object getDelegatedProvided() throws SOFAException {
		return this;
	}


	/** Store the next element in the delegation chain. */
	public void setDelegatedRequired(Object ref) throws SOFAException {		
		endpoint = (MILifecycle) ref;
	}

	/**
	 * {@link MILifecycle#exit()}
	 * Doesn't notify the invoker.
	 */
	public void exit() {		
		endpoint.exit();
	}
	
	/**
	 * {@link MILifecycle#getState()}
	 * Doesn't notify the invoker.
	 */
	public LifecycleState getState() {
		return endpoint.getState();
	}

	/**
	 * {@link MILifecycle#start()()}
	 * Notifies the invoker before the next delegation chain element is called.
	 * Tries to call the 'start' function on the scripting engine. 
	 * If the function is not found, the error is ignored.
	 */
	public void start() {				
		
		try { 
			scriptInvoker.invoke("start", null);
		} catch (javax.script.ScriptException e) {
			throw new RuntimeException(e);
		} catch (NoSuchMethodException e) {
			/* ignore */ 
		} catch (NotScriptException e) {
			/* ignore */ 
		} catch (SOFAException e) {
			throw new RuntimeException(e);
		}		
		
		endpoint.start();
	}

	/**
	 * {@link MILifecycle#stop()}
	 * Notifies the invoker before the next delegation chain element is called.
	 * Tries to call the 'stop' function on the scripting engine. 
	 * If the function is not found, the error is ignored.
	 */
	public void stop() {
		try { 
			scriptInvoker.invoke("stop", null);
		} catch (javax.script.ScriptException e) {
			throw new RuntimeException(e);
		} catch (NoSuchMethodException e) {
			/* ignore */ 
		} catch (NotScriptException e) {
			/* ignore */ 
		} catch (SOFAException e) {
			throw new RuntimeException(e);
		}		
		
		endpoint.stop();
	}
	
	/**
	 * {@link MILifecycle#waitStopped()()}
	 * Doesn't notify the invoker.
	 */
	public void waitStopped() {		
		endpoint.waitStopped();
	}


}
