/* $Id$
 *
 * SOFA 2.0 component system
 * Copyright (C) 2006-2010  DDDS, Charles University in Prague
 *                          http://d3s.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.sofa.aspects.script;


import java.io.IOException;
import java.util.Map;

import javax.script.Invocable;
import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import javax.script.ScriptException;

import org.apache.log4j.Logger;
import org.objectweb.dsrg.sofa.SOFAClassLoader;
import org.objectweb.dsrg.sofa.SOFAException;
import org.objectweb.dsrg.sofa.SOFAUpdatable;
import org.objectweb.dsrg.sofa.microarchitecture.ComponentDescription;
import org.objectweb.dsrg.sofa.microarchitecture.LifecycleState;
import org.objectweb.dsrg.sofa.microarchitecture.MILifecycle;
import org.objectweb.dsrg.sofa.microarchitecture.SOFAMicroComponent;
import org.objectweb.dsrg.sofa.microarchitecture.factory.GenericContent;
import org.objectweb.dsrg.sofa.repository.model.Architecture;
import org.objectweb.dsrg.sofa.repository.model.CodeBundle;

/**
 * This class implements the central microcomponent of the script aspect.
 * 
 * <p>Contains the scripting engine. The main task of this microcomponent is to
 * invoke the calls of the provided interfaces on the scripting engine. For this
 * purpose it implements the {@link MIScriptInvoker} microinterface.</p>
 * 
 * <p>It is also responsible for lifecycle management of a scripted component,
 * therefore it also implements the {@link MILifecycle} microinterface.</p>
 * 
 * <p>Moreover, it implements the {@link MIScript} interface to handle the requests from
 * the associated Script control interface.</p>
 * 
 * <p>This microcomponent also handles the dynamic component content updates.
 * Therefore it implements the {@link MIContentListener} interface to track the
 * content updates and the {@link SOFAUpdatable} interface to be able to
 * transfer the component's state from/scripted component.</p>
 * 
 * @author Jaroslav Keznikl
 */
public class MScriptInvoker implements MIScriptInvoker, MIScript,
		SOFAMicroComponent, MIContentListener, SOFAUpdatable {
	
	final static private Logger logger = Logger.getLogger(MScriptInvoker.class);
	
	/** The associated script engine manager. */
	ScriptEngineManager scriptEngineMgr = new ScriptEngineManager();
	
	/** 
	 * The current scripting engine.
	 * All calls are performed on this engine. 
	 * It is replaced when the content is updated.
	 */
	ScriptEngine engine = null;
	
	/** The code bundle of the associated component architecture. */
	CodeBundle codeBundle;
	
	/** References to all required interfaces as a map name->reference. */
	Map<String, Object> requiredIfaces;
	
	/** Current implementation language. */
	String language;
	
	/** The original script content. For purposes of {@link #reset()}. */
	String scriptContent;
	
	/** True if the associated component is scripted. */
	boolean isScripted = false;
	
	

	/**
	 * Tries to load the given state into the current script context using
	 * 'loadState' script function.
	 * If the function is not found, the error is ignored.
	 * @param storage The stored state to be loaded.
	 */
	public void loadState(Map<String, Object> storage) throws Throwable {
		if (engine == null)
			throw new NullPointerException("Script engine not initialized.");
		try {
			invokeFunction("loadState", storage);
		} catch (NoSuchMethodException e) {
			// ignore
		}
	}
	
	/**
	 * Tries to save the state of the current script context using
	 * 'saveState' script function.
	 * If the function is not found, the error is ignored.
	 * @param storage The storage where the state should be saved.
	 */
	public void saveState(Map<String, Object> storage) throws Throwable {
		if (engine == null)
			throw new NullPointerException("Script engine not initialized.");
		try {
			invokeFunction("saveState", storage);
		} catch (NoSuchMethodException e) {
			// ignore
		}
	}
		
	
	/**
	 * Invokes the given function with the given arguments on the scripting engine.
	 * @param method	name of the function to execute
	 * @param params	parameters of the function call
	 * @return	value returned by the script
	 * @throws NoSuchMethodException if the given method was not defined by the script
	 * @throws ScriptException	if the script execution failed
	 * @throws NotScriptException if the associated component is not a scripted component 
	 * @throws SOFAException if the script engine is not initialized
	 * @see #invokeFunction(String, Object...)
	 */
	public Object invoke(String method, Object[] params) throws SOFAException,
		NotScriptException, ScriptException, NoSuchMethodException {
		
		if (!isScripted)
			throw new NotScriptException("Component is not a script component");
		
		if (engine == null)
			throw new SOFAException("Script engine not initialized.");
		
		Object retval = null;
		
		retval = invokeFunction(method, params);			
		
		return retval;
	}

	/**
	 * Initializes this component. Stores the code bundle reference. Downloads
	 * the contents of the script implementation file. Creates and initializes
	 * the scripting engine using
	 * {@link #initScriptingEngine(String, String, CodeBundle, Map)}.
	 */
	public void init(ComponentDescription component, Map<String, String> parameters) throws SOFAException {
		isScripted = (component.getComponentContent() != null) && (component.getComponentContent() instanceof GenericContent);
		if (!isScripted)
			return;
		
		Architecture arch = component.getComponentArchitecture();
		codeBundle = arch.getCodeBundle();
		language = codeBundle.getLanguage();		
		String fileName = ScriptHelper.createScriptName(arch.getImplementation());		
		
		try {
			scriptContent = ScriptHelper.downloadScriptCode(codeBundle, fileName);			
		} catch (IOException e) {
			throw new SOFAException(e);
		}		
		
		if (isScripted) {
			GenericContent content = (GenericContent) component.getComponentContent();
			requiredIfaces = content.getRequiredInterfaces();
			
			engine = initScriptingEngine(language, scriptContent, codeBundle, requiredIfaces);
			content.registerUpdateHandler(this);
		}
	}
	
	
	
	/** 
	 * Returns itself as provided interface for interfaces 'invoker', 'script',
	 *  'lifecycle' and 'content-listener. Otherwise throws {@link SOFAException}.
	 * @param name name of the requested provided interface
	 * @return this reference
	 * @throws SOFAException if the requested interface is not implemented by this microcomponent
	 */
	public Object getProvided(String name) throws SOFAException {
		if(!name.equals("invoker") && !name.equals("script") 
				&& !name.equals("lifecycle") && !name.equals("content-listener")) {
			throw new SOFAException("The microcomponent does not provide interface named "
					+ ((name == null) ? "" : name));
		}
		return this;
	}

	/** 
	 * There are no required interfaces. All calls are ignored.
	 */
	public void setRequired(String name, Object ref) throws SOFAException {		
		// we silently ignore all calls
	}


	/**
	 * {@link MIContentListener#contentAboutToBeChanged(Object)}
	 * 
	 * Resets all cached information concerning the component content.
	 * The scripting engine is replaced by a newly initialized engine.
	 * 
	 * @param content	the new component content
	 */
	public void contentAboutToBeChanged(Object content) {
		boolean isNewScripted = (content != null) && (content instanceof GenericContent);
		
		if (isScripted) {
			logger.trace("Update: Destroying old content");
			isScripted = false;
			engine = null;
			codeBundle = null;
			language = null;
			scriptContent = null;
			requiredIfaces = null;
		} 

		if (isNewScripted) {
			logger.trace("Update: Creating new content");
			GenericContent c = (GenericContent) content;
			Architecture arch = c.getArchitecture();
			codeBundle = arch.getCodeBundle();
			language = codeBundle.getLanguage();		
			String fileName = ScriptHelper.createScriptName(arch.getImplementation());		
			
			try {
				scriptContent = ScriptHelper.downloadScriptCode(codeBundle, fileName);			
			} catch (IOException e) {
				throw new RuntimeException(e);
			}		
			
			requiredIfaces = c.getRequiredInterfaces();
				
			try {
				engine = initScriptingEngine(language, scriptContent, codeBundle, requiredIfaces);
				logger.trace("Update: setting update handler");
				c.registerUpdateHandler(this);
				isScripted = true;
			} catch (SOFAException e) {
				throw new RuntimeException(e);
			}
		}
		logger.trace("Update: Update finished");		
	}

	/**
	 * Initializes the scripting engine according to the given parameters. 
	 * 
	 * It uses the {@link EngineInitializer} implementation to create and
	 * initialize the engine.
	 * 
	 * @param language	the target scripting language
	 * @param scriptCode	the script code which implements the component architecture
	 * @param codeBundle	the bundle of the associated component architecture
	 * @param requiredIfaces	references to all required interfaces as a map name->reference
	 * @return 	newly created and initialized scripting engine
	 * @throws SOFAException
	 */
	ScriptEngine initScriptingEngine(String language, String scriptCode, CodeBundle codeBundle, Map<String, Object> requiredIfaces) throws SOFAException {
		
		EngineInitializerManager mgr = new EngineInitializerManager(scriptEngineMgr);
		EngineInitializer initializer = mgr.getInitializerByLanguage(language);
		
		if (initializer == null)
			throw new SOFAException("Scripting engine couldn't be created: no engine initializer for language "+ language);
		
		ClassLoader scriptClassLoader = new ScriptClassLoader(SOFAClassLoader.getSOFAClassLoader(), codeBundle);
		
		ScriptEngine e = initializer.createScriptingEngine(scriptClassLoader, scriptCode, codeBundle, requiredIfaces);
		
		if (!(e instanceof Invocable))
			throw new SOFAException("The new engine is not an instance of the Invocable interface");
		
		return e;
	}
	
	/**
	 * Invokes the given function with the given arguments
	 * @param name	the name of the function to be invoked
	 * @param args	the arguments of the call
	 * @return	the script code return value
	 * @throws ScriptException	if the script execution failed
	 * @throws NotScriptException	if the associated component is not a scripted component 
	 */
	Object invokeFunction(String name, Object... args) throws ScriptException, NoSuchMethodException {
		Invocable i = (Invocable) engine;			
		return i.invokeFunction(name, args);
	}

	/**
	 * {@link MIScript#evaluateScriptText(String)}
	 */
	public void evaluateScriptText(String text) throws SOFAException {
		try {
			checkEngine();
			engine.eval(text);
		} catch (ScriptException e) {
			throw new SOFAException("Evaluation of the script text failed.", e);
		}				
	}
	
	/**
	 * {@link MIScript#reset()}
	 */
	public void reset() throws SOFAException {
		if (isScripted) {
			engine = initScriptingEngine(language, scriptContent, codeBundle, requiredIfaces);
		} else {
			throw new SOFAException("Not a scripted component.");
		}
	}
	
	/**
	 * {@link MIScript#getValue(String)}
	 */
	public Object getValue(String name) throws SOFAException {
		checkEngine();
		
		return engine.get(name);
	}

	/**
	 * {@link MIScript#setValue(String, Object)}
	 */
	public void setValue(String name, Object value) throws SOFAException {
		checkEngine();
		
		engine.put(name, value);
	}
	
	/** 
	 * Checks whether the component is a script component and the engine is initialized.
	 * Otherwise it throws a {@link SOFAException}.
	 * @throws SOFAException if the component is not a scripted component or the engine is null
	 */
	void checkEngine() throws SOFAException {
		if (!isScripted)
			throw new SOFAException("Component is not a script component");
		
		if (engine == null)
			throw new SOFAException("Script engine not initialized.");
	}

	
	
	
}

