/* $Id$
 *
 * SOFA 2.0 component system
 * Copyright (C) 2006-2010  DDDS, Charles University in Prague
 *                          http://d3s.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.sofa.aspects.script;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Calendar;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.zip.CRC32;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import org.apache.log4j.Logger;
import org.objectweb.dsrg.sofa.SOFAException;
import org.objectweb.dsrg.sofa.microarchitecture.ComponentDescription;
import org.objectweb.dsrg.sofa.microarchitecture.SOFAMicroComponent;
import org.objectweb.dsrg.sofa.microarchitecture.SOFAMicroInterceptor;
import org.objectweb.dsrg.sofa.microarchitecture.factory.GenericContent;
import org.objectweb.dsrg.sofa.repository.CodeBundleHelper;
import org.objectweb.dsrg.sofa.repository.model.Architecture;
import org.objectweb.dsrg.sofa.repository.model.CodeBundle;
import org.objectweb.dsrg.sofa.repository.model.DataStreamType;

/**
 * Implementation of the {@link MIUpdateScript} control interface.
 * 
 * @author Jaroslav Keznikl 
 */
public class MUpdateScriptImpl implements SOFAMicroComponent,
		SOFAMicroInterceptor, MIUpdateScript, MIContentListener {

	final static private Logger logger = Logger.getLogger(MUpdateScriptImpl.class);
	
	/** Reference to the Script control interface. */
	MIScript scriptIfaceImpl;
	
	/** The value of the script value as it was before all dynamic updates. */
	String originalScript;
	
	/** History of script updates. */
	List<String> currentScriptHistory;
	
	/** The code bundle of the associated component architecture. */
	CodeBundle codeBundle;
	
	/** The file name of the associated component architecture script implementation. */
	String fileName;
	
	/** True if the associated component is a scripted component. */
	boolean isScripted = false;

	/** Initialize internal structures according to the given component architecture.
	 * @param arch The target component architecture.
	 * @throws SOFAException If loading of the script implementation fails.
	 */
	void init(Architecture arch) throws SOFAException {
		codeBundle = arch.getCodeBundle();
		fileName = ScriptHelper.createScriptName(arch.getImplementation());
		currentScriptHistory = new LinkedList<String>();

		try {
			InputStream is = codeBundle.getDataInputStream(
					DataStreamType.ORIGINAL, fileName, true);

			ByteArrayOutputStream output = new ByteArrayOutputStream();

			int length;
			byte[] buffer = new byte[1024];
			while ((length = is.read(buffer)) != -1) {
				output.write(buffer, 0, length);
			}

			originalScript = output.toString();
			currentScriptHistory.add(originalScript);
		} catch (IOException e) {
			throw new SOFAException(e);
		}
	}
	
	/**
	 * Initialize new microcomponent.
	 */
	public void init(ComponentDescription component,
			Map<String, String> parameters) throws SOFAException {

		isScripted = ScriptHelper.isScriptedComponent(component.getComponentContent());
		
		if (!isScripted)
			return;
		
		init(component.getComponentArchitecture());
	}	
	
	
	/** 
	 * Reset all cached code and component information. 
	 */
	public void contentAboutToBeChanged(Object content) {		
		isScripted = ScriptHelper.isScriptedComponent(content);
		
		if (!isScripted)
			return;
		try {
			init(((GenericContent)content).getArchitecture());
		} catch (SOFAException e) {
			throw new RuntimeException(e);			
		}
	}
	
	/**
	 * Store current script code into Repository.
	 * 
	 * The current script is stored under the original version.
	 * 
	 * The associated code bundle is downloaded to a temp file.
	 * Its contents are copied into a new code bundle and the implementation
	 * file is replaced with the new contents.
	 * The updated code bundle is then uploaded back into Repository under the
	 * same version as the original.
	 */
	public void storeCurrentScript() throws SOFAException {
		try {
			
			logger.trace("UpdateScript: Storing code of component to repository.");
			
			File oldFile  = File.createTempFile(codeBundle.getName(), ".jar");
			File newFile  = File.createTempFile(codeBundle.getName(), "_new.jar");
			
			logger.trace("New JAR:\n" + newFile.getAbsolutePath());
			
			CodeBundleHelper.downloadFile(oldFile, codeBundle);
			
			String script = getCurrentScript();
			
			InputStream input = new FileInputStream(oldFile);
			OutputStream output = new FileOutputStream(newFile);
			ZipInputStream inZip = new ZipInputStream(input);
			ZipOutputStream outZip = new ZipOutputStream(output);

						
			ZipEntry in;
			while ((in = inZip.getNextEntry()) != null) {
				ZipEntry out = null;
				if (in.getName().equals(fileName)) {
					byte[] bytes = script.getBytes();
					
					out = (ZipEntry) in.clone();
					out.setMethod(ZipEntry.STORED);
					out.setSize(bytes.length);
					out.setCompressedSize(bytes.length);
					
					CRC32 crc = new CRC32();
					crc.update(bytes);

					out.setCrc(crc.getValue());
					out.setTime(Calendar.getInstance().getTimeInMillis());
					
					outZip.putNextEntry(out);
					ScriptHelper.copy(new ByteArrayInputStream(script.getBytes()), outZip);
				} else {				
					// just copy "in"
					outZip.putNextEntry(in);
					ScriptHelper.copy(inZip, outZip);
				}
				inZip.closeEntry();
				outZip.closeEntry();
			}

			inZip.close();
			outZip.close();
			
			logger.trace("Uploading");
			CodeBundleHelper.uploadFile(newFile, codeBundle);
			logger.trace("Written");
		} catch (Exception e) {
			throw new SOFAException(e);
		}
	}
	
	
	
	/** 
	 * {@link MIUpdateScript#evaluateScriptText(String, Boolean)}
	 */
	public void evaluateScriptText(String text, Boolean append) throws SOFAException {
		scriptIfaceImpl.evaluateScriptText(text);
		if (!append)
			currentScriptHistory.clear();
		
		currentScriptHistory.add(text);		
	}

	/** 
	 * {@link MIUpdateScript#evaluateScriptText(String)}
	 */
	public void evaluateScriptText(String text) throws SOFAException {
		evaluateScriptText(text, true);
	}

	/** 
	 * {@link MIUpdateScript#reset()}
	 */
	public void reset() throws SOFAException {
		scriptIfaceImpl.reset();
		currentScriptHistory.clear();
	}

	/** 
	 * {@link MIUpdateScript#getValue(String)}
	 */
	public Object getValue(String name) throws SOFAException {
		return scriptIfaceImpl.getValue(name);
	}

	/** 
	 * {@link MIUpdateScript#setValue(String, Object)}
	 */
	public void setValue(String name, Object value) throws SOFAException {
		scriptIfaceImpl.setValue(name, value);
	}

	/** 
	 * {@link MIUpdateScript#getCurrentScript()}
	 */
	public String getCurrentScript() {
		StringBuilder sb = new StringBuilder();
		boolean first = true;
		String conjunction = "\n";
		for (String item : currentScriptHistory) {
			if (first)
				first = false;
			else
				sb.append(conjunction);
			sb.append(item);
		}
		return sb.toString();
	}

	/** 
	 * {@link MIUpdateScript#getOriginalScript()}
	 */
	public String getOriginalScript() {
		return originalScript;
	}

	/** 
	 * Returns itself as provided interface for interface 'content-listener'.
	 * Otherwise throws {@link SOFAException}.
	 * @param name name of the requested provided interface
	 * @return this reference
	 * @throws SOFAException if the requested interface is not implemented by this microcomponent
	 */
	public Object getProvided(String name) throws SOFAException {
		if(!name.equals("content-listener") && !name.equals("update-script")) {
			throw new SOFAException("The microcomponent does not provide interface named "
					+ ((name == null) ? "" : name));
		}
		return this;
	}

	/**
	 * Sets required the interface (script). 
	 */
	public void setRequired(String name, Object ref) throws SOFAException {
		if (name != null && name.equals("script")) {
			if (ref instanceof MIScript) {
				scriptIfaceImpl = (MIScript) ref;
			} else {
				throw new SOFAException(
						"Unable to set required interface 'script'. Incompatible type.");
			}
		}
		// we silently ignore interfaces we dont need
	}
	
	public Object getDelegatedProvided() throws SOFAException {
		return this;
	}

	/** 
	 * The microcomponent doesn't delegate any required interfaces as it's
	 * an control interface implementation.
	 */
	public void setDelegatedRequired(Object ref) throws SOFAException {
		// ignore incoming calls
	}


	
}
