/* $Id$
 *
 * SOFA 2.0 component system
 * Copyright (C) 2006-2010  DDDS, Charles University in Prague
 *                          http://d3s.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.sofa.aspects.script;

import java.io.StringReader;
import java.util.Map;

import javax.script.Invocable;
import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;

import org.objectweb.dsrg.sofa.SOFAClassLoader;
import org.objectweb.dsrg.sofa.SOFAException;
import org.objectweb.dsrg.sofa.repository.model.CodeBundle;

/**
 * {@link EngineInitializer} implementation for the Jython, the Python engine.
 * 
 * @author Jaroslav Keznikl 
 */
public class PythonEngineInitializer extends AbstractEngineInitializer implements EngineInitializer {

	/** Command to import the sys module. */
	private static final String IMPORT_SYS = "import sys";
	/** Command to import the os module. */
	private static final String IMPORT_OS = "import os";
	/** Name of the variable which stores the current class loader. */
	private static final String CLASSLOADER_VAR_NAME = "sys.classLoader";
	/** Name of the variable for the python-specific script imported. */
	private static final String PYTHON_IMPORTER_VAR_NAME = "SOFAPythonImporter";
	
	/** The name of the supported language. */
	public static final String LANGUAGE_NAME = "python";
	
	/** The associated script engine manager. */
	private ScriptEngineManager scriptEngineMgr;
	
	/** Initializes the inner fields. */
	public PythonEngineInitializer(ScriptEngineManager mgr) {
		scriptEngineMgr = mgr;
	}

	/**
	 * {@link EngineInitializer#createScriptingEngine(ClassLoader, String, CodeBundle, Map)}
	 * 
	 * <p>The class loader is configured using the {@value #CLASSLOADER_VAR_NAME}
	 * Python variable.</p>
	 * 
	 * <p>The engine is initialized with the Python-specific script module importer
	 * instance ({@link SOFAPythonImporter}). This importer is accessible via
	 * the {@value PYTHON_IMPORTER_VAR_NAME} Python variable.</p>
	 */
	public ScriptEngine createScriptingEngine(ClassLoader loader,
			String scriptCode, CodeBundle codeBundle,
			Map<String, Object> requiredIfaces) throws SOFAException {
		        
		ScriptEngine engine = scriptEngineMgr.getEngineByName(LANGUAGE_NAME);
		
		
		if (engine == null)
			throw new SOFAException("Scripting engine couldn't be created.");
		
		try {
			engine.eval(IMPORT_SYS);
			engine.eval(IMPORT_OS);
			engine.put(CLASSLOADER_VAR_NAME,loader);
			
			SOFAPythonImporter pythonImporter = new SOFAPythonImporterImpl(engine, codeBundle);			
			engine.put(PYTHON_IMPORTER_VAR_NAME, pythonImporter);
			
			SOFAScriptImporter scriptImporter = new SOFAScriptImporterImpl(engine, codeBundle);
			engine.put(SCRIPT_IMPORTER_VAR_NAME, scriptImporter);
			
			engine.eval(new StringReader(scriptCode));
		} catch (Exception e) {
			throw new SOFAException(e);
		}	
		
		for (String requiredName: requiredIfaces.keySet()) {
			engine.put(requiredName, requiredIfaces.get(requiredName));
		}
		
		return engine;
	}

}
