/* $Id$
 *
 * SOFA 2.0 component system
 * Copyright (C) 2006-2010  DDDS, Charles University in Prague
 *                          http://d3s.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.sofa.aspects.script;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import javax.script.ScriptEngine;

import org.apache.log4j.Logger;
import org.objectweb.dsrg.sofa.SOFAClassLoader;
import org.objectweb.dsrg.sofa.SOFAException;
import org.objectweb.dsrg.sofa.deployment.RepositoryErrorException;
import org.objectweb.dsrg.sofa.repository.RepositoryAgent;
import org.objectweb.dsrg.sofa.repository.model.CodeBundle;
import org.objectweb.dsrg.sofa.repository.model.DataStreamType;

/**
 * The {@link SOFAPythonImporter} implementation for the Jython 2.2.1 interpreter.
 * 
 * @author Jaroslav Keznikl
 */
public class SOFAPythonImporterImpl implements SOFAPythonImporter {
	/** The associated scripting engine. */
	ScriptEngine engine;
	
	/** The current code bundle. */
	CodeBundle cbundle;
		
	SOFAClassLoader loader = SOFAClassLoader.getSOFAClassLoader();
	final static Logger logger = Logger.getLogger(SOFAPythonImporter.class);
	
	/** Default tag to be used in {@link #loadCodeBundle(String)}. */
	private static final String CURRENT_TAG = "current";

	/** Initializes private fields. */
	SOFAPythonImporterImpl(ScriptEngine engine, CodeBundle cb) {
		this.engine = engine;
		this.cbundle = cb;
	}
	
	/**
	 * {@link SOFAPythonImporter#loadAllDependencies()}
	 */
	public void loadAllDependencies() throws IOException, SOFAException{
		loadCodeBundle(cbundle);
		
		for (CodeBundle cb : ScriptHelper.getDependencies(cbundle)) {
			loadCodeBundle(cb);
		}
	}
	
	/**
	 * {@link SOFAPythonImporter#loadCurrentCodeBundle()}
	 */
	public void loadCurrentCodeBundle() throws IOException, SOFAException{
		loadCodeBundle(cbundle);
	}
	
	/**
	 * {@link SOFAPythonImporter#loadCodeBundle(String)}
	 */
	public void loadCodeBundle(String cbName) throws IOException, SOFAException {
		loadCodeBundle(cbName, null, CURRENT_TAG);
	}	
	
	/**
	 * {@link SOFAPythonImporter#loadCodeBundle(String, String, String)}
	 */
	public void loadCodeBundle(String cbName, String cbVersion, String cbTag) throws IOException, SOFAException {
		RepositoryAgent repAgent;
		CodeBundle cb;
		
		try {
			repAgent = RepositoryAgent.init();
		} catch (IOException e) {
			throw new RepositoryErrorException ("Cannot connect to repository!", e);
		}
		
		cb = ScriptHelper.getCodeBundle(repAgent, cbName, cbVersion, cbTag);
				
		if (!checkDependency(cb, cbundle))
			throw new SOFAException("The specified code bundle " + cbName + " is not a dependency of the current code bundle " + cbundle.getName());
		
		loadCodeBundle(cb);
	}

	/**
	 * Checks, whether the @a dependency bundle depends on the @a dependent bundle.
	 * @param dependency	the dependency bundle
	 * @param dependent		the dependent bundle
	 * @return true if the dependency bundle depends on the dependent bundle.
	 */
	private boolean checkDependency(CodeBundle dependency, CodeBundle dependent) {
		
		if (equalBundles(dependency, dependent))
			return true;
		
		for (CodeBundle dep : dependent.getDependencies()) {
			if (checkDependency(dependency, dep)) {
				return true;
			}
		}
		
		return false;		
	}

	/**
	 * Checks whether the given bundles are equal.
	 * (have same name, version etc.)
	 * @param cb1	the first bundle to compare
	 * @param cb2	the second bundle to compare
	 * @return	true if the given bundles are equal
	 */
	private boolean equalBundles(CodeBundle cb1, CodeBundle cb2) {
		return cb2.getName().equals(cb1.getName()) && cb2.getVersion().toString().equals(cb1.getVersion().toString());
	}
	
	/**
	 * Loads the given code bundle in the script engine's system path, so that the engine can
	 * search this bundle for Python modules.
	 * @param cb	the target code bundle
	 * @throws IOException
	 * @throws SOFAException
	 */
	private void loadCodeBundle(CodeBundle cb) throws IOException, SOFAException {
		logger.trace("Loading code bundle: " + cb.getName());
		
		InputStream is = cb.getDataInputStream(DataStreamType.ORIGINAL);
		File tmpJar = File.createTempFile(cb.getName()+ cb.getVersion(), ".jar");
		
		logger.trace("Code bundle stored into temp JAR file: " + tmpJar.getAbsolutePath());
		
		FileOutputStream os = new FileOutputStream(tmpJar);		
		ScriptHelper.copy(is, os);		
		os.close();	
		
		try {		
			String script = "sys.path.insert(0,r\"" + tmpJar.getAbsolutePath() + "\")";
			logger.trace("Eval script: " + script);
			engine.eval(script);
		} catch (Exception e) {
			throw new SOFAException("The target code bundle " + cb.getName() + " in version " + cb.getVersion() + " couldn't be loaded!", e);
		}
		
		logger.trace("Loading code bundle finished");
	}

}
