/* $Id$
 *
 * SOFA 2.0 component system
 * Copyright (C) 2006-2010  DDDS, Charles University in Prague
 *                          http://d3s.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.sofa.aspects.script;

import java.io.IOException;
import java.io.StringReader;
import java.util.List;

import javax.script.ScriptEngine;
import javax.script.ScriptException;

import org.apache.log4j.Logger;
import org.objectweb.dsrg.sofa.SOFAClassLoader;
import org.objectweb.dsrg.sofa.SOFAException;
import org.objectweb.dsrg.sofa.deployment.RepositoryErrorException;
import org.objectweb.dsrg.sofa.repository.RepositoryAgent;
import org.objectweb.dsrg.sofa.repository.model.CodeBundle;

/**
 * Generic {@link SOFAScriptImporter} implementation.
 * 
 * @author Jaroslav Keznikl
 */
public class SOFAScriptImporterImpl implements SOFAScriptImporter {

	/** The asspciated scripting engine. */
	ScriptEngine engine;
	/** The current code bundle. */
	CodeBundle cbundle;
	
	SOFAClassLoader loader = SOFAClassLoader.getSOFAClassLoader();
	final static Logger logger = Logger.getLogger(SOFAScriptImporterImpl.class);
	
	/** Default tag to be used in {@link #importFile(String, String)}. */
	private static final String CURRENT_TAG = "current";
	
	/** Initializes private fields. */
	SOFAScriptImporterImpl(ScriptEngine engine, CodeBundle cb) {
		this.engine = engine;
		this.cbundle = cb;
	}	
	
	/**
	 * {@link SOFAScriptImporter#importFile(String)}
	 */
	public void importFile(String fileName) throws SOFAException {
		// search only stand-alone bundles
		List<CodeBundle> bundles = ScriptHelper.getDependencies(cbundle, true);
		bundles.add(0, cbundle);
		
		logger.trace("Budnles to search:" + bundles);
		
		boolean found = false;
		for (CodeBundle codeBundle : bundles) {
			logger.trace("Searching for file " + fileName + " in code bundle " + codeBundle.getName());
			try {
				found = true;
				importFile(fileName, codeBundle);
				break;
			} catch(IOException e) {
				logger.trace("IOException in importFile for file " + fileName + " and code bundle " + codeBundle.getName());				
				found = false;
				continue;
			} catch (ScriptException e) {
				throw new SOFAException("Importing of the file " + fileName + " failed!", e);
			}
			
			
		}
		
		if (!found)
			throw new SOFAException("The file " + fileName + " was not found!");
		
		
	}

	/**
	 * {@link SOFAScriptImporter#importFromCurrent(String)}
	 */
	public void importFromCurrent(String fileName) throws SOFAException  {
		try {
			importFile(fileName, cbundle);
		} catch (Exception e) {
			throw new SOFAException(e);
		}
	}
	
	/**
	 * {@link SOFAScriptImporter#importFile(String, String)}
	 */
	public void importFile(String fileName, String codeBundle) throws SOFAException  {
		importFile(fileName, codeBundle, null, CURRENT_TAG);
	}

	/**
	 * {@link SOFAScriptImporter#importFile(String, String, String, String)}
	 */
	public void importFile(String fileName, String codeBundle, String cbVersion, String cbTag) throws SOFAException  {
		RepositoryAgent repAgent;
		try {
			repAgent = RepositoryAgent.init();
		} catch (IOException e) {
			throw new SOFAException ("Cannot connect to repository!", e);
		}
		
		CodeBundle cb =  ScriptHelper.getCodeBundle(repAgent, codeBundle, cbVersion, cbTag);
		try {
			importFile(fileName, cb);
		} catch (Exception e) {
			throw new SOFAException("Importing of file " + fileName + " from code bundle " +codeBundle + " failed!", e);
		}
	}
	
	/**
	 * Import the given script file from the given code bundle.	
	 * @param fileName	the relative file path in the given code bundle
	 * @param cb	the target code bundle
	 * @throws IOException
	 * @throws ScriptException
	 */
	private void importFile(String fileName, CodeBundle cb) throws IOException, ScriptException {
		logger.trace("Getting script code file " + fileName + " from code bundle " + cb.getName());
		System.out.println("Getting script code file " + fileName + " from code bundle " + cb.getName());
		String scriptCode = ScriptHelper.downloadScriptCode(cb, fileName);
		System.out.println("Evaluating script file.");
		logger.trace("Evaluating script file.");
		engine.eval(new StringReader(scriptCode));
	}

}
