/* $Id$
 *
 * SOFA 2.0 component system
 * Copyright (C) 2006-2010  DDDS, Charles University in Prague
 *                          http://d3s.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.sofa.aspects.script;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;
import org.objectweb.dsrg.sofa.SOFAClassLoader;
import org.objectweb.dsrg.sofa.repository.model.CodeBundle;

/**
 * A custom class loader which allows the scripting engines to locate renamed
 * classes in all dependent code bundles.
 * 
 * Classes are loaded using the standard {@link SOFAClassLoader}. This class loader
 * ensures the code bundle dependency traversing.
 * 
 * @author Jaroslav Keznikl
 */
public class ScriptClassLoader extends ClassLoader {

	final static Logger logger = Logger.getLogger("ScriptClassLoader");
	
	/**
	 * The root code bundle of the dependency tree which is searched for renamed
	 * classes.
	 */
	CodeBundle bundle;
	
	/**
	 * Standard SOFA class loader instance used for loading a class from a
	 * particular code bundle.
	 */
	SOFAClassLoader loader;
	
	
	public ScriptClassLoader(SOFAClassLoader loader, CodeBundle bundle) {
		this.bundle = bundle;
		this.loader = loader;
	}

	public ScriptClassLoader(SOFAClassLoader loader, CodeBundle bundle, ClassLoader parent) {
		super(parent);
		this.bundle = bundle;
		this.loader = loader;
	}

	/**
	 * Loads the given class using the {@link SOFAClassLoader} from the given
	 * code bundle or some of its dependencies.
	 * 
	 * @param bundle	The target code bundle.
	 * @param className	Name of the class to be loaded.
	 * @return	The Class instance corresponding to the givn name and code bundle.
	 * @throws ClassNotFoundException
	 */
	Class<?> loadClass(CodeBundle bundle, String className) throws ClassNotFoundException {
		List<CodeBundle>  bundles = ScriptHelper.getDependencies(bundle);
		bundles.add(0, bundle);	
		
		logger.trace("Dependencies for bundle: " + bundle.getName() + " are: "+ bundles.toString());
		
		for (CodeBundle codeBundle : bundles) {
			logger.trace("ScriptClassLoaderHelper: Loading class from the code bundle: " + codeBundle.getName());
			Class result = null;
			try {
				result = loader.loadClass(codeBundle, className);
				return result;
				
			} catch(ClassNotFoundException e) {} // keep trying			
		}
		
		logger.trace("No suitable CodeBundle found for class: "+className);
		throw new ClassNotFoundException("No suitable CodeBundle found for class: "+className);	
	}

	/**
	 * Loads the given class.
	 * 
	 * If the target class is not found using the standard
	 * {@link SOFAClassLoader}, the loader tries to search the current code
	 * bundle and all its dependencies ({@link #loadClass(CodeBundle, String)}.
	 */
	public Class<?> loadClass(String className) throws ClassNotFoundException {

		logger.trace("ScriptClassLoaderHelper: Request for class " + className);
		
		if (loader == null)
			throw new ClassNotFoundException("No SOFAClassLoader.");

		Class result;
		try {
			result = loader.loadClass(className);
			logger.trace("ScriptClassLoaderHelper: Class "
							+ className
							+ " loaded by SOFAClassLoader without specifying the code bundle.");
			return result;
		} catch (ClassNotFoundException e) {
			if (bundle != null) {
				logger.trace("Class " + className + " was not found by default, trying to use code bundle");
				// search the code bundle and all dependencies for the given class name
				return loadClass(bundle, className);
			} else {
				throw new ClassNotFoundException("No default CodeBundle.");
			}
		}
	}
	

}
