/* $Id$
 *
 * SOFA 2.0 component system
 * Copyright (C) 2006-2010  DDDS, Charles University in Prague
 *                          http://d3s.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.sofa.aspects.script;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import javax.script.ScriptEngineManager;

import org.eclipse.emf.common.util.URI;
import org.objectweb.dsrg.sofa.deployment.RepositoryErrorException;
import org.objectweb.dsrg.sofa.microarchitecture.factory.GenericContent;
import org.objectweb.dsrg.sofa.repository.RepositoryAgent;
import org.objectweb.dsrg.sofa.repository.SofaURIFactory;
import org.objectweb.dsrg.sofa.repository.model.CodeBundle;
import org.objectweb.dsrg.sofa.repository.model.DataStreamType;

/**
 * Singleton class that contains utility methods used throughout the scripting
 * support implementation.
 * 
 * @author Jaroslav Keznikl 
 */
public final class ScriptHelper {
	
	/** Bytecode description codes of all primitive types. */
	private static Map<Class,Character> descCodes;
	
	/**
	 *  The only instance used to initialize the {@link #descCodes} map. 
	 */
	private static ScriptHelper instance = new ScriptHelper(); 
		
	/**
	 * Engine initializer used to determine supported implementation languages.
	 */
	private static EngineInitializerManager initializerManager = new EngineInitializerManager(new ScriptEngineManager());
	
	private static final int IO_BUFFER_SIZE = 4 * 1024;
	
	/**
	 * Initializes the inner fields.
	 */
	private ScriptHelper() {
		descCodes = new Hashtable<Class, Character>();
		
		descCodes.put(int.class, 'I');
		descCodes.put(long.class, 'J');
		descCodes.put(byte.class, 'B');
		descCodes.put(boolean.class, 'Z');
		descCodes.put(char.class, 'C');
		descCodes.put(float.class, 'F');
		descCodes.put(double.class, 'D');
		descCodes.put(void.class, 'V');
	}
	
	/**
	 * Returns all dependent code bundles of the given bundle.
	 * The bundles are ordered in the DFS order.
	 * @param bundle the target code bundle.
	 * @return all dependent code bundles of the given bundle.
	 */
	public static List<CodeBundle>  getDependencies(CodeBundle bundle) {
		return getDependencies(bundle, false);
	}

	/**
	 * Returns all dependent code bundles of the given bundle.
	 * The bundles are ordered in the DFS order.
	 * @param bundle the target code bundle.
	 * @param standaloneOnly	return only stand-alone bundles.
	 * @return all dependent code bundles of the given bundle.
	 */
	public static List<CodeBundle>  getDependencies(CodeBundle bundle, boolean standaloneOnly) {
		List<CodeBundle> bundles = new ArrayList<CodeBundle>();
		
		if (bundle == null)
			return bundles;
		
		for (CodeBundle codeBundle : bundle.getDependencies()) {
			if ((!standaloneOnly) || codeBundle.isStandalone())
				bundles.add(codeBundle);			
			
			List<CodeBundle> deps = getDependencies(codeBundle, standaloneOnly);			
			for (CodeBundle d : deps) {
				if (!bundles.contains(d))
					bundles.add(d);
			}
		}	
		return bundles;
	}
	
	/**
	 * Convert the package-like file description to standard file path. 
	 * @param impl the package-like file description.
	 * @return the associated file path.
	 */
	public static String createScriptName(String impl) {
		String fileName = impl.substring(0, impl.lastIndexOf('.')); 
		String extension = impl.substring(impl.lastIndexOf('.'));
		
		return fileName.replace('.', '/') + extension;		
	}
	
	/**
	 * Downloads contents of the given file from the given code bundle from the Repository
	 * @param cb 	the target code bundle.
	 * @param fileName	the target file name.
	 * @return contents of the given file from the given code bundle.
	 * @throws IOException
	 */
	public static String downloadScriptCode(CodeBundle cb, String fileName) throws IOException {
		InputStream is = cb.getDataInputStream(DataStreamType.ORIGINAL,
				fileName, true);

		ByteArrayOutputStream output = new ByteArrayOutputStream();

		int length;
		byte[] buffer = new byte[1024];
		while ((length = is.read(buffer)) != -1) {
			output.write(buffer, 0, length);
		}
		
		return output.toString();
	}
	
	/**
	 * Returns the code bundle reference associated with the given arguments.
	 * 
	 * Either version or tag of the target code bundle must not be null.
	 * 
	 * @param agent 	the repository agent.
	 * @param cbName	target code bundle name.
	 * @param cbVersion	target code bundle version.
	 * @param cbTag		target code bundle tag.
	 * @return associated code bundle reference.
	 * @throws RepositoryErrorException
	 */
	public static CodeBundle getCodeBundle(RepositoryAgent agent,
			String cbName, String cbVersion, String cbTag)
			throws RepositoryErrorException {

		URI uri = SofaURIFactory.createURI(cbName, cbVersion, cbTag);
		System.out.println(uri.toString());

		try {
			return (CodeBundle) agent.getQuery().getEObject(uri, false);
		} catch (Exception e) {
			throw new RepositoryErrorException("The code bundle named " + cbName + " in version " + cbVersion + " and tag " + cbTag + " does not exist in the repository!", e);
		}
	}
	
	/**
	 * Returns true if the given component content is a scripted component content.
	 * @param content the target component content.
	 * @return true if the given component content is a scripted component content.
	 */
	public static boolean isScriptedComponent(Object content) {
		if ((content != null) && (content instanceof GenericContent)) {
			GenericContent gc = (GenericContent) content;
			return initializerManager.supportsLanguage(gc.getArchitecture().getCodeBundle().getLanguage());
		} else {
			return false;
		}
	}
	
	/**
	 * Return byte-code signature representation of the given method.
	 * @param m method description
	 * @return byte-code signature representing the given method 
	 */
	public static String methodToDesc(Method m) {
		StringBuffer desc = new StringBuffer("(");
		
		Class c[] = m.getParameterTypes();
		
		for(int i = 0; i < c.length; ++i) {
			desc.append(classToDesc(c[i]));
		}
		
		desc.append(')');
		desc.append(classToDesc(m.getReturnType()));
		
		return desc.toString();
	}
	
	
	/**
	 * Return byte-code representation of the given type.
	 * @param c type
	 * @return byte-code representation of the type <code>c</code>.
	 */
	public static String classToDesc(Class c) {
		if(c.isPrimitive()) {
			return descCodes.get(c).toString();
		}
		
		if(c.isArray()) {
			return c.getName().replace('.', '/');
		}
		
		return "L" + c.getName().replace('.', '/') + ";";
	}
	
	/**
     * Copy the content of the input stream into the output stream, using a temporary
     * byte array buffer whose size is defined by {@link #IO_BUFFER_SIZE}.
     *
     * @param in The input stream to copy from.
     * @param out The output stream to copy to.
     *
     * @throws IOException If any error occurs during the copy.
     */
	public static void copy(InputStream in, OutputStream out)
			throws IOException {
		byte[] b = new byte[IO_BUFFER_SIZE];
		int read;
		while ((read = in.read(b)) != -1) {
			out.write(b, 0, read);
		}
	}
	
}
