/* $Id$
 *
 * SOFA 2.0 component system
 * Copyright (C) 2006-2010  DDDS, Charles University in Prague
 *                          http://d3s.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.sofa.aspects.script;

import java.lang.reflect.*;

import org.apache.log4j.Logger;
import org.objectweb.asm.*;
import org.objectweb.asm.Type;
import org.objectweb.dsrg.sofa.SOFAClassLoader;
import org.objectweb.dsrg.sofa.SOFAException;
import org.objectweb.dsrg.sofa.bootstrap.InterceptorController;
import org.objectweb.dsrg.sofa.bootstrap.InterceptorGeneratorHelper;
import org.objectweb.dsrg.sofa.bootstrap.MIInterceptor;
import org.objectweb.dsrg.sofa.microarchitecture.SOFAMicroGenerator;
import org.objectweb.dsrg.sofa.microarchitecture.SOFAMicroInterceptor;
import org.ow2.dsrg.sofa.aspects.script.NotScriptException;

import org.ow2.dsrg.sofa.aspects.script.MIScriptInterceptor;

import java.util.*;

/**
 * Generator of script interceptors.
 * 
 * Generated interceptors have the following structure:
 * 
 * public class SampleScriptInterceptor implements SampleInterface {
 *     MIScriptInterceptor listener;
 *     SampleInterface target;
 *     public int func(int a, String s) {
 *         int ret = 0;
 *         int contentRet = 0;
 *         boolean useContent = false;
 *         
 *         try {
 *             ret = ((Integer) listener.notifyCall("func", (Integer) a, s)).intValue();
 *         } catch (NotScriptException e) {
 *             useContent = true;
 *         }
 *         contentRet = target.func(a, s);
 *         if (useContent)
 *             ret = contentRet;
 *         return ret;
 *     }
 * }
 * 
 * @see MIScriptInterceptor
 * 
 * @author Jaroslav Keznikl
 */
public class ScriptInterceptorGenerator implements SOFAMicroGenerator {
	final static private Logger logger = Logger.getLogger(ScriptInterceptorGenerator.class);
	final static private SOFAClassLoader loader = SOFAClassLoader.getSOFAClassLoader();
	
	/**
	 * Hashmap of tuples (primitive_class, opcode).
	 */
	static Map<Class,Character> descCodes;
	
	/**
	 * Constructor.
	 */
	public ScriptInterceptorGenerator() {
		descCodes = new Hashtable<Class, Character>();
		
		descCodes.put(int.class, 'I');
		descCodes.put(long.class, 'J');
		descCodes.put(byte.class, 'B');
		descCodes.put(boolean.class, 'Z');
		descCodes.put(char.class, 'C');
		descCodes.put(float.class, 'F');
		descCodes.put(double.class, 'D');
		descCodes.put(void.class, 'V');
	}
	
	
	/**
	 * Generate interceptor class for given interface and parameters.
	 * 
	 * @param iface interface which would be intercepted
	 * @param parameters parameters for the bytecode generator which notification should be generated 
	 */
	public SOFAMicroInterceptor generate(Class iface, Map<String,String> parameters) throws SOFAException {
		String interceptorName = "";
		
		boolean notifyCall = parameters.containsKey("notifyCall");
		boolean notifyReturn = parameters.containsKey("notifyReturn");
		
		if(notifyCall == notifyReturn) {
			if(notifyCall){
				interceptorName = "Interceptor";
			}
			else {
				interceptorName = "DummyInterceptor";
			}
		}
		else {
			if(notifyCall) {
				interceptorName = "CallInterceptor";
			}
			else {
				interceptorName = "ReturnInterceptor";
			}
		}
		
		// Prepare string constants
		String delegated = iface.getName().replace('.', '/');
		String className = "org/objectweb/dsrg/sofa/bootstrap/Script"+interceptorName+"$"+delegated;
		
		try {
			Class cl = loader.loadClass(className.replace('/', '.'));
			logger.debug("One is loaded, we are finished.");
			return new InterceptorController(cl.newInstance());
		}
		catch (ClassNotFoundException e) {}	// not loaded, proceed and generate one
		catch (Exception e) {
			throw new SOFAException("Error while instantiating interceptor.", e);
		}

		
		// Let's dance !               
		ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS);
		generateHeader(cw, className, delegated);
		generateConstructor(cw, className, delegated);
	
		// generate methods from the interface
		Method[] methods = iface.getMethods();
		for(int i = 0; i < methods.length; ++i) {
			Method method = methods[i];
			
			generateMethod(cw, className, delegated, method, notifyCall, notifyReturn);
		}
			
		// save the result class into a file if system property is set
		if (InterceptorGeneratorHelper.saveInterceptors()) {		
			InterceptorGeneratorHelper.dumpClass(className.replace('/', '.'), cw);
		}
		
		// instantiate and store to cache
		try {
			byte[] b = cw.toByteArray();
			InterceptorGeneratorHelper.dumpClass("MujInterceptor", cw);
			Class cl = loader.loadClass(b, className);
			return new InterceptorController(cl.newInstance());
		}
		catch (Exception e) {
			throw new SOFAException("Error while instantiating generated interceptor.", e);
		}
	}
	
	/**
	 * Generate a class header.
	 * 
	 * @param cw class writter
	 * @param className name of generated class
	 * @param delegated name of delegated interface
	 */
	void generateHeader(ClassWriter cw, String className, String delegated) {
		String delegatedDesc = "L"+delegated+";";
		String ifaces[] = { delegated };
		
		cw.visit(49, 4145, className, null, "java/lang/Object", ifaces);
		cw.visitField(Opcodes.ACC_PUBLIC, "target", delegatedDesc, null, null);
		cw.visitField(Opcodes.ACC_PUBLIC, "listener", classToDesc(MIScriptInterceptor.class), null, null);
	}
	
	/**
	 * Generate a non-parametric constructor.
	 * 
	 * @param cw class writter
	 * @param className name of generated class
	 * @param delegated name of delegated interface
	 */
	void generateConstructor(ClassWriter cw, String className, String delegated) {
		MethodVisitor mv = cw.visitMethod(Opcodes.ACC_PUBLIC, "<init>", "()V", null, null);
		mv.visitVarInsn(Opcodes.ALOAD, 0);
		mv.visitMethodInsn(Opcodes.INVOKESPECIAL, "java/lang/Object", "<init>", "()V");
		mv.visitInsn(Opcodes.RETURN);
		mv.visitMaxs(1,1);
	}
	
	/**
	 * Generate method header and implementation.
	 *  
	 * @param cw class writter
	 * @param className name of generated class
	 * @param delegated name of delegated interface
	 * @param method description of generated method
	 * @param notifyCall if it is true generator will generate notification code on the call of this method
	 * @param notifyReturn if it is true generator will generate notification code on the return from this method
	 */
	void generateMethod(ClassWriter cw, String className, String delegated, Method method, boolean notifyCall, boolean notifyReturn) throws SOFAException {
		
		String methodName = method.getName().trim();
		String methodDesc = methodToDesc(method);
		String returnType = classToDesc(method.getReturnType());

		Class param[] = method.getParameterTypes();
		Integer maxlocals = Math.max(1, param.length + 1); // count also this reference
		Integer maxstack = 2;
		Integer retValuePos = 0;
		Integer tmpValuePos = 0;
		
		Label afterContentCall = new Label();
		
		MethodVisitor mv = cw.visitMethod(Opcodes.ACC_PUBLIC, methodName, methodDesc, null, null);
		
		// if function is not void 
		if (classToChar(method.getReturnType()) != 'V') { 
			maxlocals+=2; // add two local variables for return value and temporary variable, if needed
			retValuePos = maxlocals-2; // ad local variable for return value
			tmpValuePos = maxlocals-1; // ad last local variable for temporary variable			
		}
		
		/* add local variable, that will store the information, 
		 * whether the content return value should be used */
		int useContentPos = maxlocals;
		maxlocals++;
		
		InstructionProvider instProvider = getInstructionProvider(method.getReturnType());
		
		/* initialize local variables containing return value */
		instProvider.initializeReturnValue(mv, retValuePos);		
		instProvider.initializeReturnValue(mv, tmpValuePos);
		
		/* Use content return value by default */
		mv.visitInsn(Opcodes.ICONST_1);
		mv.visitVarInsn(Opcodes.ISTORE, useContentPos);
		
		if(notifyCall){ // notifyCall			
			
			/* don't use content return value if calling script aspect */
			mv.visitInsn(Opcodes.ICONST_0);
			mv.visitVarInsn(Opcodes.ISTORE, useContentPos);
			
			String notifyType = MIScriptInterceptor.class.getName().replace('.', '/');

			/* Declare labels for exception handling */
			Label exceptionStart = new Label();
			Label exceptionEnd = new Label();
			Label exceptionHandler = new Label();
			Label exceptionAfter = new Label();
			
			String exceptionType = NotScriptException.class.getCanonicalName().replace('.', '/'); // "scriptaspect/NotScriptException";
			
			mv.visitTryCatchBlock(exceptionStart, exceptionEnd, exceptionHandler, exceptionType);
			mv.visitLabel(exceptionStart);
			
			/* Notify listener */
			mv.visitVarInsn(Opcodes.ALOAD, 0);
			//mv.visitInsn(Opcodes.DUP); // duplicate this pointer
			mv.visitFieldInsn (Opcodes.GETFIELD, className, "listener", classToDesc(MIScriptInterceptor.class));
			mv.visitLdcInsn (methodName+methodDesc);
			mv.visitLdcInsn (delegated);
			
			/* Copy formal parameters and call converter method */
			copyParametersToArray(mv, param, maxstack, maxlocals);
			
			/* Correct operands should be on the stack:  this, methodName, operands as Object[] */
			mv.visitMethodInsn(Opcodes.INVOKEINTERFACE, notifyType,	"notifyCall", "(Ljava/lang/String;Ljava/lang/String;[Ljava/lang/Object;)Ljava/lang/Object;");
			instProvider.convertAndStore(mv, retValuePos);
			
			mv.visitLabel(exceptionEnd);
			mv.visitJumpInsn(Opcodes.GOTO, exceptionAfter);
			

			/* Exception handler */
			
			mv.visitLabel(exceptionHandler);
			/* in catch block local variables will be the same and stack will contain just the reference on the exception */
			mv.visitFrame(Opcodes.F_SAME1, 0, null, 1, new Object[] {exceptionType});
			mv.visitVarInsn(Opcodes.ASTORE, maxlocals); //store the exception into local variable
			
			/* ignore the incoming exception */
			/* and mark, that return value of content should be used */
			mv.visitInsn(Opcodes.ICONST_1);
			mv.visitVarInsn(Opcodes.ISTORE, useContentPos);	
			
			
			/* Decide, whether to call content implementation 
			 * If notifyCall finished successfuly, than the content return value won't be used.
			 */		
			mv.visitLabel(exceptionAfter);
			/* after try-catch block the local variables and stack will be the same as before  */
			mv.visitFrame(Opcodes.F_SAME, 0, null, 0, null);
						
			
		}
		
		/* Load target class address */
		mv.visitVarInsn(Opcodes.ALOAD, 0);
		mv.visitFieldInsn (Opcodes.GETFIELD, className, "target", "L"+delegated+";" );
		/* Load parameteres */
		copyParameters(mv, param, maxstack, maxlocals);
		
		/* Call the delegation chain */
		mv.visitMethodInsn(Opcodes.INVOKEINTERFACE, delegated, methodName, methodDesc);
		/* Store and convert the returned Object */
		instProvider.store(mv, tmpValuePos);
		
		/* If the content return value shouldn't be used skip 
		 * the assignment of the temporary value */
		mv.visitVarInsn(Opcodes.ILOAD, useContentPos);
		mv.visitJumpInsn(Opcodes.IFEQ, afterContentCall);
		
		/* Store the content return value stored in the temporary variable
		 * into the variable which will be returned */
		instProvider.load(mv, tmpValuePos);
		instProvider.store(mv, retValuePos);		
		
		mv.visitLabel(afterContentCall);
		mv.visitFrame(Opcodes.F_SAME, 0, null, 0, null);
		
		if(notifyReturn){ // notifyReturn
			/* notify return is ignored for script interceptor */
		}

		/* Copy return value */
		instProvider.generateReturnStatement(mv, retValuePos);
		
		maxstack  = Math.max(maxstack, maxlocals);
		mv.visitMaxs(maxstack, maxlocals);
	}
	
	/**
	 * Return byte-code representation of method signature.
	 * 
	 * @param m method description
	 * 
	 * @return byte-code representing method signature
	 */
	static String methodToDesc(Method m) {
		StringBuffer desc = new StringBuffer("(");
		
		Class c[] = m.getParameterTypes();
		
		for(int i = 0; i < c.length; ++i) {
			desc.append(classToDesc(c[i]));
		}
		
		desc.append(')');
		desc.append(classToDesc(m.getReturnType()));
		
		return desc.toString();
	}
	
	/**
	 * Return byte-code representation of a given type.
	 * 
	 * @param c type
	 * @return byte-code representation of the type <code>c</code>.
	 */
	static String classToDesc(Class c) {
		if(c.isPrimitive()) {
			return descCodes.get(c).toString();
		}
		
		if(c.isArray()) {
			return c.getName().replace('.', '/');
		}
		
		return "L" + c.getName().replace('.', '/') + ";";
	}
	
	/**
	 * Returns a symbol representing given type in byte-code.	  
	 * 
	 * @param c type
	 * @return a symbol representing the type <code>c</code> in byte-code
	 */
	static char classToChar(Class c) {
		if(c.isPrimitive()) {
			return descCodes.get(c);
		}
		
		if(c.isArray()) {
			return '[';
		}
		
		return 'L';
	}

	/**
	 * Copy parameters of a method to stack as array of objects.
	 *	
	 * TODO pouziva se bitpush, tzn je omezeny pocet argumentu na velikost byte.
	 * 
	 * 
	 * @param mv method visitor
	 * @param param an array of method parameters
	 */
	void copyParametersToArray (MethodVisitor mv, Class param[], Integer maxstack, Integer maxlocals) {
		int paramCount = param.length;
		int pos = 0;
		
		// allocate new Object array for parameter values
		mv.visitIntInsn(Opcodes.BIPUSH, paramCount);
		mv.visitTypeInsn(Opcodes.ANEWARRAY, "java/lang/Object");
		
		for(int j = 0; j < paramCount; ++j) {
			char c = classToChar(param[j]);
			++pos;
			
			logger.debug("Storing parameter to array: "+param[j].getName());
			
			switch(c) {
			case 'L':  // object reference
			case '[':  // array reference
				
				mv.visitInsn(Opcodes.DUP); // load array pointer
				mv.visitIntInsn(Opcodes.BIPUSH, pos - 1); // target position in array
				mv.visitVarInsn(Opcodes.ALOAD, pos); // parameter to be stored into the array
				mv.visitInsn(Opcodes.AASTORE); // store the parameter
				
				break;
			case 'I':   // int
			case 'B':   // byte
			case 'Z':   // boolean
			case 'C':	// char
			case 'F':   // float
			case 'D':   // double
			case 'J':   // long			
				mv.visitInsn(Opcodes.DUP); // load array pointer
				mv.visitIntInsn(Opcodes.BIPUSH, pos - 1); // target position in array				
				
				// transform the primitive type value on the stack into the Object instance
				switch (c) {
				case 'I':   // int
					mv.visitVarInsn(Opcodes.ILOAD, pos); // load parameter to be stored into the array
					mv.visitMethodInsn(Opcodes.INVOKESTATIC, "java/lang/Integer", "valueOf", "(I)Ljava/lang/Integer;");
					break;
				case 'B':   // byte
					mv.visitVarInsn(Opcodes.ILOAD, pos); 
					mv.visitMethodInsn(Opcodes.INVOKESTATIC, "java/lang/Byte", "valueOf", "(B)Ljava/lang/Byte;");
					break;
				case 'Z':   // boolean
					mv.visitVarInsn(Opcodes.ILOAD, pos);
					mv.visitMethodInsn(Opcodes.INVOKESTATIC, "java/lang/Boolean", "valueOf", "(Z)Ljava/lang/Boolean;");
					break;
				case 'C':	// char
					mv.visitVarInsn(Opcodes.ILOAD, pos);
					mv.visitMethodInsn(Opcodes.INVOKESTATIC, "java/lang/Character", "valueOf", "(C)Ljava/lang/Character;");
					break;
				case 'F':   // float
					mv.visitVarInsn(Opcodes.FLOAD, pos);
					mv.visitMethodInsn(Opcodes.INVOKESTATIC, "java/lang/Float", "valueOf", "(F)Ljava/lang/Float;");
					break;
				case 'D':   // double
					mv.visitVarInsn(Opcodes.DLOAD, pos);
					mv.visitMethodInsn(Opcodes.INVOKESTATIC, "java/lang/Double", "valueOf", "(D)Ljava/lang/Double;");
					break;
				case 'J':   // long
					mv.visitVarInsn(Opcodes.LLOAD, pos);
					mv.visitMethodInsn(Opcodes.INVOKESTATIC, "java/lang/Long", "valueOf", "(J)Ljava/lang/Long;");
					break;
				}
								
				mv.visitInsn(Opcodes.AASTORE); // store the parameter
				break;
			default:
				logger.error("ScriptInterceptorGenerator: Unhandled parameter type " + param[j].getName());
				break;
			}
		}
	}
	
	
	/**
	 * Copy parameters of a method to byte-code call of delegated method.
	 * 
	 * @param mv method visitor
	 * @param param an array of method parameters
	 * @param maxstack 
	 * @param maxlocals 
	 * 
	 * @see MethodVisitor 
	 */
	void copyParameters (MethodVisitor mv, Class param[], Integer maxstack, Integer maxlocals){
		int paramCount = param.length;
		int pos = 0;
		
		for(int j = 0; j < paramCount; ++j) {
			char c = classToChar(param[j]);
			++pos;
			
			logger.debug("Loading parameter: "+param[j].getName());
			
			switch(c) {
			case 'L':  // object reference
			case '[':  // array reference
				mv.visitVarInsn(Opcodes.ALOAD, pos);
				break;

			case 'I':  // integer
			case 'B':  // byte
			case 'Z':  // boolean
			case 'C':  // char
				mv.visitVarInsn(Opcodes.ILOAD, pos);
				break;
				
			case 'F':  // float
				mv.visitVarInsn(Opcodes.FLOAD, pos);
				break;
				
			case 'D':  // double
				mv.visitVarInsn(Opcodes.DLOAD, pos);
				++pos;
				++maxlocals;
				break;

			case 'J':  // long
				mv.visitVarInsn(Opcodes.LLOAD, pos);
				++pos;
				++maxlocals;
				break;

			default:
				logger.error("Unhandled parameter type " + param[j].getName());
				break;
			}
		}
	}
	
	
	
	/**
	 * Interface for type-specific bytecode operations.
	 */
	private interface InstructionProvider {
		
		/**
		 * Generate byte-code for initialization of local variable containing return value.
		 * @param mv	the method visitor
		 * @param position	the target stack position
		 * @see MethodVisitor
		 */
		void initializeReturnValue(MethodVisitor mv, Integer position);
		
		/**
		 * Generate byte-code for storing return value into local variable at the given position.
		 * @param mv	the method visitor
		 * @param position	the target stack position
		 * @see MethodVisitor
		 */
		void store(MethodVisitor mv, Integer position);
		
		/**
		 * Generate byte-code for loading the return value from local variable at the given position.
		 * @param mv	the method visitor
		 * @param position	the target stack position
		 * @see MethodVisitor
		 */
		void load(MethodVisitor mv, Integer position);
		
		/**
		 * Generate byte-code for converting the return value from Object reference
		 * and store it into local variable at given position.
		 * @param mv	the method visitor
		 * @param position	the target stack position
		 * @see MethodVisitor
		 */
		void convertAndStore(MethodVisitor mv, Integer position);
		
		/**
		 * Generate byte-code return statement.
		 * @param mv	the method visitor
		 * @param position	the target stack position
		 * @see MethodVisitor
		 */
		void generateReturnStatement(MethodVisitor mv, Integer position);		
	}
	
	/** {@link InstructionProvider} implementation for void type. */
	private class VoidInstructionProvider implements InstructionProvider {		
		public void generateReturnStatement(MethodVisitor mv, Integer position) {
			mv.visitInsn(Opcodes.RETURN);
		}
		public void initializeReturnValue(MethodVisitor mv, Integer position) {
			/* ignore */			
		}
		public void store(MethodVisitor mv, Integer position) {
			/* ignore */
		}
		public void load(MethodVisitor mv, Integer position) {
			/* ignore */
		}
		public void convertAndStore(MethodVisitor mv, Integer position) { 
			/* ignore */
			// remove the returned Object value of the notifyCall function
			mv.visitInsn(Opcodes.POP);
		}
	}
	
	/** {@link InstructionProvider} implementation for Object type. */
	private class ObjectInstructionProvider implements InstructionProvider {
		Class targetClass;
		
		ObjectInstructionProvider(Class targetClass) {
			this.targetClass = targetClass;
		}
		
		public void generateReturnStatement(MethodVisitor mv, Integer position) {
			mv.visitVarInsn(Opcodes.ALOAD, position);
			mv.visitInsn(Opcodes.ARETURN);
		}
		public void initializeReturnValue(MethodVisitor mv, Integer position) {
			mv.visitInsn(Opcodes.ACONST_NULL);
			mv.visitVarInsn(Opcodes.ASTORE, position);
		}
		public void store(MethodVisitor mv, Integer position) {
			mv.visitVarInsn(Opcodes.ASTORE, position);
		}
		public void load(MethodVisitor mv, Integer position) {
			mv.visitVarInsn(Opcodes.ALOAD, position);
		}
		public void convertAndStore(MethodVisitor mv, Integer position) {
			mv.visitTypeInsn(Opcodes.CHECKCAST, targetClass.getName().replace('.', '/'));
			mv.visitVarInsn(Opcodes.ASTORE, position); 
		}
	}
	
	/** {@link InstructionProvider} implementation for int type. */
	private class IntegerInstructionProvider implements InstructionProvider {		
		public void generateReturnStatement(MethodVisitor mv, Integer position) {
			mv.visitVarInsn(Opcodes.ILOAD, position);
			mv.visitInsn(Opcodes.IRETURN);
		}
		public void initializeReturnValue(MethodVisitor mv, Integer position) {
			mv.visitInsn(Opcodes.ICONST_0);			
			mv.visitVarInsn(Opcodes.ISTORE, position);
		}
		public void store(MethodVisitor mv, Integer position) {
			mv.visitVarInsn(Opcodes.ISTORE, position);			
		}
		public void load(MethodVisitor mv, Integer position) {
			mv.visitVarInsn(Opcodes.ILOAD, position);			
		}
		public void convertAndStore(MethodVisitor mv, Integer position) { 
			Label nullLabel = new Label();

			mv.visitTypeInsn(Opcodes.CHECKCAST, "java/lang/Integer");
			mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/lang/Integer", "intValue", "()I");			
			mv.visitVarInsn(Opcodes.ISTORE, position);
		}
	}
	
	/** {@link InstructionProvider} implementation for float type. */
	private class FloatInstructionProvider implements InstructionProvider {		
		public void generateReturnStatement(MethodVisitor mv, Integer position) {
			mv.visitVarInsn(Opcodes.FLOAD, position);
			mv.visitInsn(Opcodes.FRETURN);
		}
		public void initializeReturnValue(MethodVisitor mv, Integer position) {
			mv.visitInsn(Opcodes.FCONST_0);			
			mv.visitVarInsn(Opcodes.FSTORE, position);
		}
		public void store(MethodVisitor mv, Integer position) {
			mv.visitVarInsn(Opcodes.FSTORE, position);			
		}
		public void load(MethodVisitor mv, Integer position) {
			mv.visitVarInsn(Opcodes.FLOAD, position);			
		}
		public void convertAndStore(MethodVisitor mv, Integer position) { 
			Label nullLabel = new Label();
			
			mv.visitTypeInsn(Opcodes.CHECKCAST, "java/lang/Float");
			mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/lang/Float", "floatValue", "()F");			
			mv.visitVarInsn(Opcodes.FSTORE, position);			
		}
	}
	
	/** {@link InstructionProvider} implementation for double type. */
	private class DoubleInstructionProvider implements InstructionProvider {		
		public void generateReturnStatement(MethodVisitor mv, Integer position) {
			mv.visitVarInsn(Opcodes.DLOAD, position);
			mv.visitInsn(Opcodes.DRETURN);
		}
		public void initializeReturnValue(MethodVisitor mv, Integer position) {
			mv.visitInsn(Opcodes.DCONST_0);			
			mv.visitVarInsn(Opcodes.DSTORE, position);
		}
		public void store(MethodVisitor mv, Integer position) {
			mv.visitVarInsn(Opcodes.DSTORE, position);			
		}
		public void load(MethodVisitor mv, Integer position) {
			mv.visitVarInsn(Opcodes.DLOAD, position);			
		}
		public void convertAndStore(MethodVisitor mv, Integer position) { 
			Label nullLabel = new Label();

			mv.visitTypeInsn(Opcodes.CHECKCAST, "java/lang/Double");
			mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/lang/Double", "doubleValue", "()D");			
			mv.visitVarInsn(Opcodes.DSTORE, position);
 
		}
	}
	
	/** {@link InstructionProvider} implementation for long type. */
	private class LongInstructionProvider implements InstructionProvider {		
		public void generateReturnStatement(MethodVisitor mv, Integer position) {
			mv.visitVarInsn(Opcodes.LLOAD, position);
			mv.visitInsn(Opcodes.LRETURN);
		}
		public void initializeReturnValue(MethodVisitor mv, Integer position) {
			mv.visitInsn(Opcodes.LCONST_0);			
			mv.visitVarInsn(Opcodes.LSTORE, position);
		}
		public void store(MethodVisitor mv, Integer position) {
			mv.visitVarInsn(Opcodes.LSTORE, position);			
		}
		public void load(MethodVisitor mv, Integer position) {
			mv.visitVarInsn(Opcodes.LLOAD, position);			
		}
		public void convertAndStore(MethodVisitor mv, Integer position) { 
			Label nullLabel = new Label();
			
			mv.visitTypeInsn(Opcodes.CHECKCAST, "java/lang/Long");
			mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/lang/Long", "longValue", "()J");			
			mv.visitVarInsn(Opcodes.LSTORE, position);
		}
	}
	
	
	/**
	 * Get instruction provider according to the return type
	 * @param mv
	 * @param retType
	 * @param position
	 */
	InstructionProvider getInstructionProvider(Class retType) throws SOFAException {
		switch(classToChar(retType)) {
		case 'V':   // void			
			/* ignore */
			return new VoidInstructionProvider();			
		case 'L':   // object reference
		case '[':   // array reference
			return new ObjectInstructionProvider(retType);
		case 'I':   // int
		case 'B':   // byte
		case 'Z':   // boolean
		case 'C':	// char			
			return new IntegerInstructionProvider();			
		case 'F':   // float						
			return new FloatInstructionProvider();
		case 'D':   // double
			return new DoubleInstructionProvider();			
		case 'J':   // long						
			return new LongInstructionProvider();			
		default:
			logger.error("Unsupported return type " + retType.getName());
			throw new SOFAException("Unsupported return type " + retType.getName());			
		}
	}
	
	
}	


